# -*- coding: utf-8 -*-

import collections
import ipaddress
import logging
import random
import re

from dateutil.relativedelta import relativedelta

from odoo import fields, models
from odoo.addons.saas_worker.config import saas_config
from odoo.addons.saas_worker.const import SPECIAL_PREFIXES, TEST
from odoo.addons.saas_worker.util import email_on_exception, get_ip_info
from odoo.exceptions import ValidationError
from odoo.http import request
from odoo.tools.safe_eval import safe_eval, wrap_module

_logger = logging.getLogger(__name__)

wrapped_re = wrap_module(
    re,
    ['findall', 'finditer', 'fullmatch', 'search', 'A', 'I', 'S', 'L', 'M', 'T', 'U', 'X']
)

TRIAL_SCOPE = [('scope', 'in', (False, 'trial'))]

class TrialRules(models.Model):
    _name = 'openerp.enterprise.trial.rules'
    _description = 'Anti-spam rules for trial creations'
    _order = "name"

    name = fields.Char('Rule Name', required=True)
    description = fields.Char('Description')
    type = fields.Selection([
        ('name', 'Name (regex)'),
        ('email', 'Email (regex)'),
        ('ip_addr', 'IP Address (CIDR accepted)'),
        ('geoip', 'GeoIP (country code)'),
        ('ip_info', 'ipinfo.io lookup (Python code)'),
    ], 'Rule Type', required=True)
    scope = fields.Selection([
        ('trial', 'Trial Databases'),
        ('dns', 'Domain Name Registrations'),
    ], 'Rule Scope', help="Applies only for this scope (not set = all scopes)", required=False)
    rule = fields.Char('Rule', required=True)
    active = fields.Boolean('Active', default=True)


    @email_on_exception(to='sys@odoo.com', cc='', reraise=False)
    def _process_trial_creation(self, **kwargs):
        """ Process anti-spam rules, regroup the rules by type first
            and process them.

            return: TrialRules's record: the blocking rule.
                    True, will allow the trial creation.
                    None, in case of exception will allow the trial creation
                        if an exception is raised, it will be caught by the
                        `email_on_exception` wrapper and warn us by email.
        """
        # if no rules passed, retrieve all the active rules (forced in the context)
        all_rules = self or self.with_context(active_test=True).search(TRIAL_SCOPE)

        # group them by rule type
        grouped_rules = collections.defaultdict(self.browse)
        for r in all_rules:
            grouped_rules[r.type] |= r

        # call the rules method processing
        for rtype, rules in grouped_rules.items():
            if hasattr(rules, f'_process_rule_{rtype}'):
                process_rule_method = getattr(rules, f'_process_rule_{rtype}')
                # call the processing method, if it fails, return the blocking rule
                res = process_rule_method(**kwargs)
                if res is not True:
                    return res
        return True

    def _process_rule_name(self, **kwargs):
        """ Note: this method naively consider the rules as name rules

            Rules are non case sensitive regex.
              e.g. (microsoft|micr0s0ft)
        """
        db_name = kwargs.get('db_name', '').lower().strip()
        if not db_name:
            return False

        for rule in self:
            # search the pattern in the db name, if found return the blocking rule
            if re.search(rule.rule, db_name, re.IGNORECASE) is not None:
                _logger.info('Trial creation rejected for "%s", matching name rule "%s" (#%s), "%s" matches "%s", ip: %s',
                    db_name,
                    rule.name,
                    rule.id,
                    db_name,
                    rule.rule.strip(),
                    request.httprequest.remote_addr if request else 'n/a',
                )
                return rule
        # no match, the db name is accepted
        return True

    def _process_rule_email(self, **kwargs):
        """ Note: this method naively consider the rules as email rules

            Rules are non case sensitive regex.
              e.g. microsoft.+@gmail.com
        """
        email = kwargs.get('email', '').lower().strip()
        if not email:
            return False

        for rule in self:
            # search the pattern in the email, if found return the blocking rule
            if re.search(rule.rule, email, re.IGNORECASE) is not None:
                _logger.info('Trial creation rejected for "%s", matching email rule "%s" (#%s), "%s" matches "%s", ip: %s',
                    kwargs.get('db_name', ''),
                    rule.name,
                    rule.id,
                    email,
                    rule.rule.strip(),
                    request.httprequest.remote_addr,
                )
                return rule
        # no match, the email is accepted
        return True

    def _process_rule_ip_addr(self, **kwargs):
        """ Note: this method naively consider the rules as ip_addr rules

            Rules are IPv4/IPv6 addresses, accepting the CIDR notation.
              e.g. 178.33.40.43
                   2001:41d0:203:61d5::
                   178.33.40.43/24
                   2001:41d0:203:61d5::/16
        """
        ipaddr = ipaddress.ip_address(kwargs.get('ip_addr') or request.httprequest.remote_addr)
        # if we match one of the blocked IP we return the blocking rule
        for rule in self:
            # disabled the strict mode to allow us to set CIDR address with the hosts bits set
            if ipaddr in ipaddress.ip_network(rule.rule, strict=False):
                _logger.info('Trial creation rejected for "%s", matching IP rule "%s" (#%s), "%s" matches "%s", ip: %s',
                    kwargs.get('db_name', ''),
                    rule.name,
                    rule.id,
                    request.httprequest.remote_addr,
                    rule.rule,
                    request.httprequest.remote_addr,
                )
                return rule
        # no IP match, then it is ok
        return True

    def _process_rule_geoip(self, **kwargs):
        """ Note: this method naively consider the rules as geoip rules

            Rules are ISO 3166-2 country code
              e.g. "BE" for Belgium
                   "US" for United States of America
        """
        country_code = request.geoip.country_code
        if not country_code:
            return True
        # if we match one of the blocked country we return the blocking rule
        for rule in self:
            if (rule.rule or '').lower() == country_code.lower():
                _logger.info('Trial creation rejected for "%s", matching GeoIP rule "%s" (#%s), "%s" matches "%s", ip: %s',
                    kwargs.get('db_name', ''),
                    rule.name,
                    rule.id,
                    country_code,
                    rule.rule,
                    request.httprequest.remote_addr,
                )
                return rule
        # no country match, then it is ok
        return True

    def _process_rule_ip_info(self, **kwargs):
        """ Note: this method naively consider the rules as ip_info rules


            Rules are python code executed using `safe_eval` with a context evaluation.
            Containing `ip_info` (see the data structure here: https://ipinfo.io/developers/data-types#asn-data)
            and access to the `re` library for regexes.

            Returning a True will block the trial creation, returning a False will allow it.

            Example:
                    To block the Autonomous System n°15169, just type the rule.

                    ip_info['asn']['asn'] == 'AS15169'

                    To block Belgium for trials using the module `website`.

                    ":website:" in trial_info["selected_app"] and ip_info['country'] == "BE"

        """
        # avoid processing ipinfo when not called on a request
        if not request:
            return True
        # retrieve the ip's information
        ip_info = get_ip_info(saas_config['accounts'].get('ipinfo_token', ''), kwargs.get('ip_addr'))

        # failed to retrieve the info for this IP, we do not block
        if ip_info.get('error'):
            _logger.info(
                'Trial creation bypasses the IPInfo checking rule for "%s" with IP address "%s". Reason: "%s"',
                kwargs.get('db_name', ''),
                request.httprequest.remote_addr,
                ip_info['error']
            )

            if random.randint(1, 20) == 1:  # raise ~5% of exception to have a email and detect long downtime / errors
                raise Exception(ip_info['error'])
            return True

        eval_context = {
            'ip_info': ip_info,
            'trial_info': kwargs,
            're': wrapped_re,
        }

        for rule in self:
            rule_res = safe_eval(rule.rule, eval_context)

            if rule_res:
                _logger.info('Trial creation rejected for "%s", matching IPInfo rule "%s" (#%s), "%s" matches "%s", ip: %s',
                    kwargs.get('db_name', ''),
                    rule.name,
                    rule.id,
                    ip_info,
                    rule.rule,
                    request.httprequest.remote_addr,
                )
                return rule

        return True

    def _check_dbname(self, name, scopes=(False,)):
        """ Check if a database name is valid or not following `name` rules

        :param tuple scopes: (optional) apply only rules for the given scopes
        """
        domain = [('scope', 'in', scopes), ('type', '=', 'name')]
        name_rules = self.with_context(active_test=True).search(domain)
        return name_rules._process_rule_name(db_name=name) is True

    def _process_trial_limit(self, **kwargs):
        """ Check if the current IP address is not created too many trials
        """
        if not request:
            return True

        # do not block trial creation on our staging servers
        if TEST:
            return True

        is_internal_user = request.env.user.has_group('base.group_user')
        stonecutter = request.httprequest.headers.get('X-Is-Stone-Cutter')
        whitelisted = request.httprequest.headers.get('X-Is-Whitelisted')
        # if one of those headers is present on the request we do not limit the trial creation
        # those headers are added on the loadbalancer side
        # Stone cutter is the one used for employees to bypass the load balancer limit.
        # We use this header to identify our employees and it is a dynamic one, we add the ip when loading a certain request
        # done automatically when loading the Odoo backend
        # The whitelist is a static list where we manually add the IP addresses
        if is_internal_user or whitelisted or stonecutter:
            return True

        # if the database name starts with a special prefix, we do not limit the trial creation
        # since the student for the edu program create multiple databases from the same IP address
        # we do not want to limit them
        db_name = kwargs.get('db_name', '').lower().strip()
        if db_name.startswith(SPECIAL_PREFIXES):
            return True

        limit_per_day = int(self.env['ir.config_parameter'].get_param('openerp_enterprise.trial_limit_per_day', '5'))
        ip_address = request.httprequest.remote_addr

        # we count how many databases were created in the last 24 hours
        trial_count = self.env['openerp.enterprise.trial.log'].search_count([
            ('ip_address', '=', ip_address),
            ('creation_date', '>=', fields.Datetime.now() - relativedelta(days=1))
        ])

        if trial_count > limit_per_day:
            _logger.info(
                'Trial creation rejected for "%s", trial limit per day reached count %d/%d, ip: %s',
                db_name,
                trial_count,
                limit_per_day,
                ip_address,
            )
            return False

        return True

    def button_test_rule(self):
        """ This method will crash if the odoo.http.request is not bound to a real HTTP Request, this is intended.

            This function is bound to a form view button that allows us to test single rules to be sure no
            syntax error has been made. It won't really help to try out complex rules as the dataset to test will
            always be the same.
        """
        self.ensure_one()

        kwargs = {
            "csrf_token": request.csrf_token(None),
            "grc3_token": "",
            "username": "John+Doe",
            "email": "john.doe@odoo.com",
            "company_name": "Section+9",
            "db_name": "section-9",
            "phone": "+32123456789",
            "country_id": "20",  # BE
            "lang": ["fr_FR","en_US"],
            "company_size": "5-20",
            "plan": "plan_to_use",
            "cuuid": "",
            "selected_app": "crm:sale_management:purchase:website:website_sale:mass_mailing_sms:hr:hr_holidays:hr_expense:lunch",
            "hosting": "test-eu3.test.odoo.com",
            "template": "",
            "verification": ""
        }

        res = "Unknown rule type!"
        try:
            if hasattr(self, f'_process_rule_{self.type}'):
                process_rule_method = getattr(self, f'_process_rule_{self.type}')
                # call the processing method, if it fails, return the blocking rule
                res = process_rule_method(**kwargs)
        except Exception as e:
            res = f"Exception: {e}"

        raise ValidationError(f'Rule test returned (openerp.enterprise.trial.rules(XXX) = Blocked, True = Good to Go):\n\n{res}')
